from __future__ import annotations

from typing import Sequence

from ..device import Device
from ..field import DefaultExtraField
from .base import GeneratorBase


class CaddyFileGenerator(GeneratorBase):
    def __init__(self, devices: Sequence[Device[DefaultExtraField]]):
        super().__init__()
        self.devices = devices

    # def generate(self):
    #     hosts = {}
    #     for dev in self.devices:
    #         hosts.setdefault(dev.group, [])
    #         hosts[dev.group].append(dev.get_name())
    #         if dev.extra.ansible is not None:
    #             for alias in dev.extra.ansible.aliases:
    #                 hosts.setdefault(alias, [])
    #                 hosts[alias].append(dev.get_name())

    #     for group, hosts in hosts.items():
    #         self.add_line(f"[{group}]")
    #         for host in hosts:
    #             self.add_line(host)
    #         self.add_line()



from __future__ import annotations

from typing import Sequence

from ..device import Device
from ..field import DefaultExtraField
from .base import GeneratorBase


class AnsibleHostsGenerator(GeneratorBase):
    def __init__(self, devices: Sequence[Device[DefaultExtraField]]):
        super().__init__()
        self.devices = devices

    def generate(self):
        group_hosts: dict[str, list] = {}
        for dev in self.devices:
            group_hosts.setdefault(dev.group, [])
            group_hosts[dev.group].append(dev.get_name())
            if dev.extra.ansible is not None:
                for alias in dev.extra.ansible.aliases:
                    group_hosts.setdefault(alias, [])
                    group_hosts[alias].append(dev.get_name())

        for group, hosts in group_hosts.items():
            self.add_line(f"[{group}]")
            for host in hosts:
                self.add_line(host)
            self.add_line()



from __future__ import annotations

import os
from sys import stdout

import toml
from loguru import logger
from pydantic import ValidationError

from .device import Device
from .field import BaseExtraField, DefaultExtraField
from .generator import AnsibleHostsGenerator, DNSConfigGenerator, SSHHostsGenerator


def get_devices[T: BaseExtraField](
    devices_config: list[dict], extra_field_cls: type[T] = BaseExtraField
) -> list[Device[T]]:
    devices = []
    for device in devices_config:
        try:
            extra_field_cls.model_validate(device.get("extra", {}))
            devices.append(Device[extra_field_cls].model_validate(device))
        except ValidationError:
            logger.warning(f"device {device} has invalid extra field")
            continue
    return devices


def generate_config(path: str = "~/.config/autoconfig/config.toml", *, log_level="INFO"):
    logger.remove()
    logger.add(stdout, level=log_level)
    path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    config = toml.load(open(path, "r"))
    devices = get_devices(config["devices"], DefaultExtraField)
    generator = AnsibleHostsGenerator(devices)
    generator.write("~/.config/ansible/hosts")
    generator = SSHHostsGenerator(devices)
    generator.write("~/.ssh/config")
    generator = DNSConfigGenerator(devices)
    generator.write("~/.config/autoconfig/auto-dns.json")



from __future__ import annotations

import json
import re
from typing import Sequence

from ..device import Device
from ..field import DefaultExtraField
from .base import GeneratorBase


class DNSConfigGenerator(GeneratorBase):
    def __init__(self, devices: Sequence[Device[DefaultExtraField]]):
        super().__init__()
        self.devices = devices

    def generate(self):
        record_list = []
        for device in self.devices:
            name = device.get_target()
            if device.extra.dns is not None:
                target = device.extra.dns.target
                if re.match(r"\b(?:\d{1,3}\.){3}\d{1,3}\b", target):
                    record_type = "A"
                else:
                    record_type = "CNAME"
            else:
                target = "0.0.0.0"
                record_type = "A"
            if device.main:
                record_list.append((device.group, "CNAME", f"{name}.bone6.top"))
            record_list.append((f"{name}", record_type, target))

        self.generated_code = json.dumps({"domain": "bone6.top", "records": record_list}, indent=2)



from __future__ import annotations

from pydantic import BaseModel, Field


class BaseExtraField(BaseModel):
    pass


class SSHField(BaseModel):
    port: int
    user: str
    forward_agent: bool = True


class AnsibleField(BaseModel):
    aliases: list[str] = Field(default_factory=list)


class DNSField(BaseModel):
    target: str


class DefaultExtraField(BaseExtraField):
    ssh: SSHField | None = None
    ansible: AnsibleField | None = None
    dns: DNSField | None = None



from __future__ import annotations

import os

from loguru import logger


class GeneratorBase:
    def __init__(self):
        self.generated_code: str | None = None

    def add_line(self, line: str = "", *, indentation: int = 0):
        if self.generated_code is None:
            self.generated_code = ""
        self.generated_code += f"{' '*indentation}{line}\n"

    def generate(self):
        raise NotImplementedError("generate() should be implemented by subclass")

    def clear(self):
        self.generated_code = None

    @property
    def code(self):
        if self.generated_code is None:
            self.generate()
            assert self.generated_code is not None, "generate() should set self.generated_code"

        return self.generated_code

    def write(self, path: str):
        path = os.path.expanduser(path)
        path = os.path.expandvars(path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        logger.info(f"writed to {path}")
        open(path, "w").write(self.code)



from __future__ import annotations

from typing import Any, TypeVar

import libcst as cst
from libcst import (
    Arg,
    AssignTarget,
    Call,
    ClassDef,
    FlattenSentinel,
    Index,
    Parameters,
    SimpleStatementLine,
    SimpleString,
    Subscript,
    SubscriptElement,
)
from libcst.codemod import (
    CodemodContext,
    VisitorBasedCodemodCommand,
    gather_files,
    parallel_exec_transform_with_prettyprint,
)
from libcst.codemod.visitors import AddImportsVisitor
from libcst.metadata import Scope, ScopeProvider


class ReplaceNodes(cst.CSTTransformer):
    def __init__(self, replacements: dict[cst.CSTNode, cst.CSTNode]):
        self.replacements = replacements

    def on_leave(self, original_node: cst.CSTNode, updated_node: cst.CSTNode):
        return self.replacements.get(original_node, updated_node)


def gen_type_param(
    type_param: cst.TypeVar | cst.TypeVarTuple | cst.ParamSpec, type_name: cst.Name | None = None
) -> SimpleStatementLine:
    """
    To generate the following code:
        T = TypeVar("T")
        P = ParamSpec("P")
        Ts = TypeVarTuple("Ts")
    """
    type_name = type_param.name if type_name is None else type_name
    match type_param:
        case cst.TypeVar(name, bound):
            args = [
                cst.Arg(SimpleString(f'"{type_name.value}"')),
            ]
            if bound is not None:
                args.append(Arg(bound, keyword=cst.Name("bound")))
            return SimpleStatementLine(
                [
                    cst.Assign(
                        targets=[AssignTarget(type_name)],
                        value=Call(
                            func=cst.Name("TypeVar"),
                            args=args,
                        ),
                    )
                ]
            )
        case cst.TypeVarTuple(name) | cst.ParamSpec(name):
            return SimpleStatementLine(
                [
                    cst.Assign(
                        targets=[AssignTarget(type_name)],
                        value=Call(
                            func=cst.Name(type_param.__class__.__name__),
                            args=[
                                Arg(SimpleString(f'"{type_name.value}"')),
                            ],
                        ),
                    )
                ]
            )
        case _:
            raise NotImplementedError


def __wrapper_func_remove_type_parameters():
    __remove_type_parameters_T = TypeVar("__remove_type_parameters_T", bound=cst.FunctionDef | ClassDef)

    def remove_type_parameters(
        node: __remove_type_parameters_T, prefix: str = "", suffix: str = ""
    ) -> tuple[list[SimpleStatementLine], __remove_type_parameters_T]:
        type_params = node.type_parameters
        if type_params is None:
            return [], node
        statements = []
        new_node = node.with_changes(type_parameters=None)

        slices = []
        for type_param in type_params.params:
            new_name = type_param.param.name.with_changes(value=f"{prefix}{type_param.param.name.value}{suffix}")
            statements.append(gen_type_param(type_param.param, new_name))
            slices.append(
                SubscriptElement(
                    slice=Index(value=new_name),
                ),
            )

        if isinstance(new_node, ClassDef):
            generic_base = Arg(
                value=Subscript(
                    value=cst.Name(
                        value="Generic",
                        lpar=[],
                        rpar=[],
                    ),
                    slice=slices,
                )
            )
            new_node = new_node.with_changes(bases=[*new_node.bases, generic_base])

        return statements, new_node

    return remove_type_parameters


remove_type_parameters = __wrapper_func_remove_type_parameters()


def gen_func_wrapper(node: cst.FunctionDef, type_vars: list[SimpleStatementLine]) -> cst.FunctionDef:
    wrapper = cst.FunctionDef(
        name=cst.Name(value=f"__wrapper_func_{node.name.value}"),
        params=Parameters(),
        body=cst.IndentedBlock(
            body=[
                *type_vars,
                node,
                SimpleStatementLine(
                    [
                        cst.Return(
                            value=cst.Name(
                                value=node.name.value,
                                lpar=[],
                                rpar=[],
                            )
                        ),
                    ]
                ),
            ]
        ),
    )
    return wrapper


class RemoveAnnotationCommand(VisitorBasedCodemodCommand):
    METADATA_DEPENDENCIES = (ScopeProvider,)

    def __init__(self, context: CodemodContext) -> None:
        self.node_to_wrapper: dict[cst.FunctionDef | ClassDef, Any] = {}
        super().__init__(context)

    def visit_FunctionDef(self, node: cst.FunctionDef):
        AddImportsVisitor.add_needed_import(self.context, "typing", "TypeAlias")
        AddImportsVisitor.add_needed_import(self.context, "typing", "TypeVar")
        AddImportsVisitor.add_needed_import(self.context, "typing", "ParamSpec")
        AddImportsVisitor.add_needed_import(self.context, "typing", "TypeVarTuple")

        type_params = node.type_parameters
        if type_params is None:
            return False

        type_scope = self.get_metadata(ScopeProvider, type_params)
        body_scope = self.get_metadata(ScopeProvider, node.body)
        replacemences = {}
        prefix = f"__{node.name.value}_"
        for type_param in type_params.params:
            for scope in [type_scope, body_scope]:
                assert isinstance(scope, Scope)
                for access in set(scope.accesses[type_param.param.name]):
                    assert isinstance(access.node, cst.Name)
                    replacemences[access.node] = cst.Name(value=f"{prefix}{type_param.param.name.value}")
        new_node = node.visit(ReplaceNodes(replacemences))
        assert isinstance(new_node, cst.FunctionDef)

        type_vars, new_node = remove_type_parameters(new_node, prefix=prefix)

        if type_vars:
            wrapper = gen_func_wrapper(new_node, type_vars)
            self.node_to_wrapper[node] = wrapper
        return False

    def leave_FunctionDef(self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef):
        wrapper = self.node_to_wrapper.get(original_node, None)
        if wrapper is None:
            return updated_node
        func = SimpleStatementLine(
            [
                cst.Assign(
                    targets=[AssignTarget(updated_node.name)],
                    value=Call(
                        func=cst.Name(value=f"__wrapper_func_{updated_node.name.value}"),
                        args=[],
                    ),
                )
            ]
        )

        return FlattenSentinel([wrapper, func])

    def visit_ClassDef(self, node: ClassDef):
        AddImportsVisitor.add_needed_import(self.context, "typing", "Generic")
        AddImportsVisitor.add_needed_import(self.context, "typing", "TypeVar")
        AddImportsVisitor.add_needed_import(self.context, "typing", "ParamSpec")
        AddImportsVisitor.add_needed_import(self.context, "typing", "TypeVarTuple")

        type_params = node.type_parameters
        if type_params is None:
            return True

        scopes = [self.get_metadata(ScopeProvider, type_params)]
        for b in node.body.body:
            scopes.append(self.get_metadata(ScopeProvider, b))
            if isinstance(b, cst.FunctionDef):
                scopes.append(self.get_metadata(ScopeProvider, b.body))

        replacemences = {}
        prefix = f"__{node.name.value}_"
        for type_param in type_params.params:
            for scope in scopes:
                assert isinstance(scope, Scope)
                for access in set(scope.accesses[type_param.param.name]):
                    assert isinstance(access.node, cst.Name)
                    replacemences[access.node] = cst.Name(value=f"{prefix}{type_param.param.name.value}")
        new_node = node.visit(ReplaceNodes(replacemences))
        assert isinstance(new_node, ClassDef)

        type_vars, new_node = remove_type_parameters(new_node, prefix=prefix)

        if type_vars:
            self.node_to_wrapper[node] = new_node, type_vars
        return True

    def leave_ClassDef(self, original_node: ClassDef, updated_node: ClassDef):
        wrapper = self.node_to_wrapper.get(original_node, None)
        if wrapper is None:
            return updated_node
        new_node, type_vars = wrapper
        return FlattenSentinel([*type_vars, new_node])


command_instance = RemoveAnnotationCommand(CodemodContext())
files = gather_files(".")
result = parallel_exec_transform_with_prettyprint(
    command_instance,
    files,
    format_code=False,
    hide_generated=True,
    # hide_progress=True,
)



from __future__ import annotations

from sys import stdout

import typer
from loguru import logger

from auto_config import utils

app = typer.Typer()


@app.command()
def generate_config(path: str = "~/.config/autoconfig/config.toml", *, log_level="INFO"):
    logger.remove()
    logger.add(stdout, level=log_level)
    utils.generate_config(path)


if __name__ == "__main__":
    app()
